/* TOC

'A			internal structures
	'A			streams
	'B			tokens
	'C			macros
'D			internal functions
	'D			streams
		'D			generic stream functions
		'F			fd stream
		'G			memory stream
	'H			token functions
	'I			parse functions
	'J			macro functions
'K			public functions
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <err.h>
#include <stdio.h>
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>

static int dofile(lua_State *L);
static int dostring(lua_State *L);
static int stacktrace(lua_State *L);

#define CLL_INSERT_BEFORE(K,Z) \
  do {  (K)->prev->next= (Z);\
        (Z)->prev= (K)->prev;\
        (Z)->next= (K);\
        (K)->prev= (Z); } while(0)

#define CLL_INIT(Y) \
  do { (Y) -> prev= (Y)->next= (Y); } while(0)


static struct {
	char *name;
	lua_CFunction f;
} entries[]= {
	{ "stacktrace", stacktrace },
	{ "m5file", dofile },
	{ "m5str", dostring },
	{ 0,0 },
};

typedef struct m5stream *M5STREAM;
ssize_t m5read(M5STREAM st,char *buf,size_t len);
M5STREAM m5stream_open(void);
ssize_t m5stream_copy(M5STREAM dst,M5STREAM src);
void m5stream_close(M5STREAM s);

lua_State *Lua;
static void exec_lua(M5STREAM d,char *x,int line,char *pn) ;
/*************************************************************************

			INTERNAL        STRUCTURES
			 - streams
			 - tokens
			 - macros

 *************************************************************************/
/* '##A' */
typedef struct xnode {
	struct xnode *prev,*next;
	void *data;
} LIST;

struct stream_ops {
	ssize_t (*write)(void *,const void *buf,size_t len);
	ssize_t (*read)(void *,void *buf,size_t len);
	void (*close)(void *);
	char* (*read_whole)(void *);
};

struct m5stream {
	void *udata;
	struct stream_ops *ops;
	int nlines;
	char *prettyname;
};

/* these are for memory stream */
typedef struct block {
	struct block *prev,*next;
	char *data;
	size_t used;
	size_t size;
} BL;

typedef struct {
	BL head;
	BL *current; /* for reading */
	size_t coff;
	int ualloc;
} MS;

static struct stream_ops nullops; /* all zeros */
static struct m5stream devnull= { 0, &nullops };
static M5STREAM nullstream= &devnull;


/* '##B' */
enum toktype { tk_identifier, tk_lparen, tk_rparen, tk_comma, tk_comment,
				tk_dollar, tk_rcurly, tk_lbegin, tk_lend,
				tk_strliteral,  tk_space, tk_other, tk_error };

typedef struct _token {
	enum toktype type;
	M5STREAM output;
} *TOKEN;

/* '##C' */


/*************************************************************************

			INTERNAL        FUNCTIONS
			 - stream functions
			 - token functions
			 - parse functions
			 - macro functions
*************************************************************************/

/* '##D'  GENERIC STREAM */

static inline ssize_t sread(M5STREAM s,char *buf,size_t len) {
	if (s->ops->read)
		return (*s->ops->read)(s->udata,buf,len);
	else
		return len;
}

static inline ssize_t m5stream_write(M5STREAM s,const char *buf,size_t len) {
	if (s->ops->write)
		return (*s->ops->write)(s->udata,buf,len);
	else
		return len;
}

void m5stream_close(M5STREAM s) {
	if (s->ops->close)
		(*s->ops->close)(s->udata);
	if (s!=nullstream)
		free(s);
}

static inline char *sread_whole(M5STREAM src) {
	if (src->ops->read_whole)
		return (*src->ops->read_whole)(src->udata);
	else
		return 0;
}

ssize_t m5stream_copy(M5STREAM dst,M5STREAM src) {
	char buf[4096];
	ssize_t rw;
	ssize_t ret= 0;
	do {
		rw= sread(src, buf,sizeof(buf));
		if (!rw)
			break;
		rw= m5stream_write(dst, buf,rw);
		ret+= rw;
	} while(rw>0);
	return ret;
}


/* '##F'   FILE STREAM */

static M5STREAM fdstream_open(int fd) {
	static struct stream_ops fdops= {
		(void*) &write,
		(void*) &read,
	};
	M5STREAM s;
        s= malloc(sizeof(*s));
	s->nlines= 0;
	s->ops= &fdops;
	s->udata= (void*) (unsigned long long) fd;
	return s;
}



/* '##G'  MEMORY STREAM */

static ssize_t memm5stream_write(MS *m, const void *_buf,size_t len) {
	BL *last= m->head.prev;
	const char *buf= _buf;
	ssize_t ret= 0;
	errno= 0;
	while(len) {
		size_t w;
		if (last==&m->head || last->used==last->size) {
			size_t osize= last->size;
			last= malloc(sizeof(*last));
			last->size= osize*2;
			last->used= 0;
			last->data= malloc(last->size);
			CLL_INSERT_BEFORE(&m->head,last);
		}
		w= len;
		if (w>last->size-last->used)
			w= last->size-last->used;
		memcpy(last->data+last->used,buf,w);
		buf+= w;
		ret+= w;
		len-= w;
		last->used+= w;
	}
	return ret;
}

static ssize_t memsread(MS *m,void *_buf,size_t len) {
	char *buf= _buf;
	ssize_t ret= 0;
	if (!m->current)
		return ret;
	while(len) {
		size_t r;
		if (m->current==&m->head || m->coff==m->current->used) {
			m->current= m->current->next;
			m->coff= 0;
			if (m->current==&m->head) {
				m->current= 0;
				break;
			}
		}
		r= len;
		if (r>m->current->used-m->coff)
			r= m->current->used-m->coff;
		memcpy(buf,m->current->data+m->coff,r);
		buf+= r;
		ret+= r;
		len-= r;
		m->coff+= r;
	}
	return ret;
}

static void memstream_close(MS *m) {
	BL *p,*n;
	for(p=m->head.next;p!=&m->head;p=n) {
		n= p->next;
		if (!m->ualloc)
			free(p->data);
		free(p);
	}
	free(m);
}

static char *memsread_whole(MS *m) {
	size_t size;
	BL *p;
	char *data;
	for(size=0,p=m->head.next;p!=&m->head;p=p->next)
		size+= p->used;
	if (!size)
		data= strdup("");
	else {
		data= malloc(size+1);
		for(size=0,p=m->head.next;p!=&m->head;p=p->next) {
			memcpy(data+size,p->data,p->used);
			size+= p->used;
		}
		data[size]= 0;
	}
	return data;
}

M5STREAM m5stream_open(void) {
	static struct stream_ops memstream_ops= {
		(void*) memm5stream_write,
		(void*) memsread,
		(void*) memstream_close,
		(void*) memsread_whole,
	};
	M5STREAM s;
	MS *m;
        m= malloc(sizeof(*m));
	CLL_INIT(&m->head);
	m->head.data= 0;
	m->current= &m->head;
	m->coff= 0;
	m->head.size= 16;
	m->ualloc= 0;
        s= malloc(sizeof(*s));
	s->nlines= 0;
	s->udata= m;
	s->ops= &memstream_ops;
	s->prettyname= 0;
	return s;
}

static M5STREAM memscopen(char *x) {
	M5STREAM s;
	BL *b;
	MS *m;
	s= m5stream_open();
	m= s->udata;
	m->ualloc= 1;
        b= malloc(sizeof(*b));
	b->used= strlen(x);
	b->size= b->used;
	b->data= x;
	CLL_INSERT_BEFORE(&m->head,b);
	return s;
}

/* '##H'  TOKEN FUNCTIONS */



static void next(TOKEN t,M5STREAM s,int *curr) {
	char c= *curr;
	ssize_t r;
	if (c=='\n')
		s->nlines++;
	m5stream_write(t->output,&c,1);
	r= sread(s,&c,1);
	if (r<=0)
		*curr= EOF;
	else
		*curr= c;
}

static TOKEN new_token(enum toktype typ) {
	TOKEN t;
	t= malloc(sizeof(*t));
	t->type= typ;
	t->output= m5stream_open();
	return t;
}

static void token_destroy(TOKEN t) {
	m5stream_close(t->output);
	free(t);
}

/* '##I'  parse functions */

static void parse(M5STREAM d,M5STREAM s);

static inline int isidentifierbegin(int x) {
	return isalpha(x) || x=='_' || x=='@';
}

static inline int isidentifier(int x) {
	return isidentifierbegin(x) || isdigit(x);
}

static TOKEN skip_char(M5STREAM s,int *curr,enum toktype typ) {
	TOKEN t= new_token(typ);
	next(t,s,curr);
	return t;
}

static void give_error(int line,char *name, char *msg) {
	fprintf(stderr, "line %d",line+1);
	if (name)
		fprintf(stderr," of %s",name);
	fprintf(stderr,": %s\n",msg);
	exit(1);
}

static TOKEN skip_c_comment(TOKEN t,M5STREAM s,int *curr) {
	int lineno= s->nlines;
	next(t,s,curr);
	while(1) {
		switch(*curr) {
			case '*':
				next(t,s,curr);
				if (*curr=='/') {
					next(t,s,curr);
					t->type= tk_comment;
					return t;
				}
				break;
			case EOF:
				give_error(lineno, s->prettyname, "C comment not closed");
				return 0;
			default:
				next(t,s,curr);
		}
	}
}

static TOKEN skip_cpp_comment(TOKEN t,M5STREAM s,int *curr) {
	int lineno= s->nlines;
	next(t,s,curr);
	while(1) {
		switch(*curr) {
			case '\n':
				next(t,s,curr);
				t->type= tk_comment;
				return t;
			case EOF:
				give_error(lineno, s->prettyname,"C++ comment not closed");
				return 0;
			default:
				next(t,s,curr);
		}
	}
}

static TOKEN skip_space(M5STREAM s,int *curr) {
	TOKEN t= new_token(tk_space);
	next(t,s,curr);
	while(*curr!=EOF && isspace(*curr))
			next(t,s,curr);
	return t;
}

static TOKEN skip_strliteral(M5STREAM s,int *curr) {
	int lineno= s->nlines;
	int q= *curr;
	TOKEN t= new_token(tk_strliteral);
	next(t,s,curr);
	while(1) {
			if (*curr==q) {
				next(t,s,curr);
				return t;
			}
			if (*curr=='\\')
				next(t,s,curr);
			if (*curr==EOF)
				break;
			next(t,s,curr);
	}
	give_error(lineno, s->prettyname, "string literal not closed");
	return 0;
}


static TOKEN skip_identifier(M5STREAM s,int *curr) {
	TOKEN t= new_token(tk_identifier);
	while(*curr!=EOF && isidentifier(*curr))
		next(t,s,curr);
	return t;
}

static TOKEN token(M5STREAM s,int *curr) {
	TOKEN t= 0;
	while(*curr!=EOF) {
		switch(*curr) {
			case '/':
				if (t) return t;
				t= new_token(tk_other);
				next(t,s,curr);
			    if (*curr=='/') return skip_cpp_comment(t,s,curr);
				else if (*curr=='*') return skip_c_comment(t,s,curr);
				break;
			case '{':
				if (t) {
					if (t->type==tk_dollar) {
						t->type= tk_lbegin;
						next(t,s,curr);
					}
					return t;
				}
				return skip_char(s,curr,tk_other);
				break;
			case '}':
				if (t)
					return t;
				t= skip_char(s,curr,tk_rcurly);
				break;
			case '$':
				if (t) {
					if (t->type==tk_rcurly) {
						t->type= tk_lend;
						next(t,s,curr);
					}
					return t;
				}
				t= skip_char(s,curr,tk_dollar);
				break;
			case '\'':
			case '"':
				return t ? t :  skip_strliteral(s,curr);
			case EOF:
				break;
			case '(':
				return t ? t : skip_char(s,curr,tk_lparen);
			case ')':
				return t ? t : skip_char(s,curr,tk_rparen);
			case ',':
				return t ? t : skip_char(s,curr,tk_comma);
			default:
				if (isidentifierbegin(*curr))
					return t ? t : skip_identifier(s,curr);
				if (isspace(*curr))
					return t ? t : skip_space(s,curr);
				if (!t)
					t= new_token(tk_other);
				next(t,s,curr);
		}
	}
	return t; /* EOF reached. could be NULL or tk_other */
}

static void compile_error() {
	errx(3,"compile error");
}

void print_stackpos(lua_State *L,int x) {
	printf("[%3d] (%10s)", x, lua_typename(L,lua_type(L,x)));
	switch(lua_type(L,x)) {
		case LUA_TNUMBER: printf("%d", (int) lua_tonumber(L,x)); break;
		case LUA_TBOOLEAN:
			printf("%s", lua_toboolean(L,x) ? "true":"false");
			break;
		case LUA_TSTRING:
			printf("\"%s\"",lua_tostring(L,x));
			break;
		case LUA_TFUNCTION:
		case LUA_TTABLE:
		case LUA_TTHREAD:
		case LUA_TUSERDATA:
		case LUA_TLIGHTUSERDATA:
			break;
	}
	printf("\n");
}

void print_stack() {
	int i;
	int n= lua_gettop(Lua);
	for(i=1;i<=n;i++)
		print_stackpos(Lua,i);
}

void m5stream_ws(M5STREAM s, const char *t) {
	m5stream_write(s, (char*) t, strlen(t));
}

void stringify(lua_State *L,M5STREAM s,int x) {
	char buf[200];
	m5stream_ws(s,"[");
	m5stream_ws(s, (char*) lua_typename(L,lua_type(L,x)));
	m5stream_ws(s,"]");
	switch(lua_type(L,x)) {
		case LUA_TNUMBER: sprintf(buf," %d", (int) lua_tonumber(L,x)); break;
		case LUA_TBOOLEAN:
			sprintf(buf," %s", lua_toboolean(L,x) ? "true":"false");
			break;
		case LUA_TSTRING:
			m5stream_ws(s, " \"");
			m5stream_ws(s, lua_tostring(L,x));
			m5stream_ws(s, "\"");
			return ;
		case LUA_TFUNCTION:
#if 0
			if (lua_iscfunction(L,x)) {
				lua_CFunction f= lua_tocfunction(L,x);
				int i;
				for(i=0;entries[i].name;i++)
					if (entries[i].f==f)
						break;
				if (entries[i].name) {
					sprintf(buf," {%s}",entries[i].name);
					break;
				}
			} else {
				char *fn;
				/* OK. x may be negative. convert it to a positive index
				   first */
				if (x<0) 
					x= lua_gettop(L)+1-x;
				lua_pushnil(L);
				fn= 0;
				while(!fn && lua_next(L,LUA_GLOBALSINDEX)!=0) {
					if (lua_type(L,-1)==LUA_TFUNCTION &&
						!lua_iscfunction(L,-1) && lua_equal(L,x,-1)) {
							fn= strdup(lua_tostring(L,-2));
					}
					lua_pop(L,1);
				}
				lua_pop(L,1);
				if (fn) {
					sprintf(buf," %s",fn);
					break;
				}
			}
                   This code used to lookup the given value in the globals table 
                   and print its name, but now it doesnt. The globalsindex stuff
                   has changed and I forgot what it was.
#endif
		case LUA_TTABLE:
		case LUA_TTHREAD:
		case LUA_TUSERDATA:
		case LUA_TLIGHTUSERDATA:
			m5stream_ws(s,"...");
		case LUA_TNIL:
			return ;
	}
	m5stream_ws(s, buf);
}

char *asstr(int i) {
	static char buf[200];
	sprintf(buf,"%d",i);
	return buf;
}

static void get_lineinfo(lua_Debug *db, char **Rfn,int *Rdef,int *Rcur) {
	if (db->source[0]=='@') {
		*Rfn= strdup(db->source+1);
		*Rdef= db->linedefined;
		*Rcur= db->currentline;
		return ;
	}
	const char *z= db->source;
	const char *ll= strstr(z,"luablock@(");
	const char *e;
	if (!ll)
		goto err;
	z= ll+strlen("luablock@(");
	for(e=z+strlen(z)-1;e!=z && *e!=')';e--)
		;
	if (e==z)
		goto err;
	for(ll=e;ll!=z && *ll!=':';ll--)
		;
	if (ll==z)
		goto err;
	int dline= strtol(ll+1,0,0);
	*Rdef= db->linedefined+dline;
	*Rcur= db->currentline+dline;
	char *fn= malloc(ll-z);
	memcpy(fn,z,ll-z);
	fn[ll-z]= 0;
	*Rfn= fn;
	return ;
err:
	*Rfn= db->short_src;
	*Rdef= db->linedefined;
	*Rcur= db->currentline;
	return ;
}

static int stacktrace(lua_State *L) {
	M5STREAM e= m5stream_open();
	{
		char *s= (char*) lua_tostring(L,-1);
		m5stream_write(e,s,strlen(s));
		lua_pop(L,1);
		m5stream_ws(e,"\n");
	}
	m5stream_ws(e,"Backtrace:\n");
	int i=1;
	lua_Debug db;
	while(lua_getstack(L,i,&db)) {
		lua_getinfo(L, "Snl", &db);
		char *fname;
		int defined,current;
		get_lineinfo(&db, &fname, &defined, &current);

		m5stream_ws(e, fname);
		m5stream_ws(e,":");
		m5stream_ws(e, asstr(current));
		m5stream_ws(e," ");
		m5stream_ws(e,db.namewhat);
		m5stream_ws(e," ");
		m5stream_ws(e, db.name ? db.name : "");
		m5stream_ws(e,"@");
		m5stream_ws(e, asstr(defined));
		m5stream_ws(e,"(");
		int j=1;
		const char *name;
		while((name=lua_getlocal(L,&db, j))) {
			if (j>1)
				m5stream_ws(e,",");
			m5stream_ws(e,name);
			m5stream_ws(e,"=");
			stringify(L,e,-1);
			lua_pop(L,1);
			j++;
		}
		m5stream_ws(e,")\n");
		i++;
	}
	char *z= sread_whole(e);
	lua_pushstring(L,z);
	free(z);
	m5stream_close(e);
	return 1;
}

void print_stackc(char *msg) {
	fprintf(stderr,"Stack Contents: %s\n",msg);
	{
		M5STREAM e= m5stream_open();
		int i;
		for(i=1;i<=lua_gettop(Lua);i++) {
			char buf[200];
			sprintf(buf,"[%d] ", i);
			m5stream_ws(e,buf);
			stringify(Lua,e,i);
			m5stream_ws(e,"\n");
		}
		char *z= sread_whole(e);
		fwrite(z, 1, strlen(z), stderr);
	}
}

static int parse_exec_macro(M5STREAM d,char *name,M5STREAM s,int *curr) {
	int lineno= s->nlines;

	if (name[0]=='@')
		name++;

	/* . */
        lua_getglobal(Lua, "all_macros");
	if (!lua_istable(Lua,-1))
        {
		lua_pop(Lua,1);
		return 0;
	}
	/* all_macros . */
	lua_pushstring(Lua,name);
	lua_gettable(Lua, -2);
	if (!lua_isfunction(Lua,-1)) {
	   lua_pop(Lua,2);
	   return 0;
	}
	lua_remove(Lua,-2);
	lua_pushcfunction(Lua, stacktrace);
	lua_pushvalue(Lua, -2);
	lua_remove(Lua,-3);

	while(1) {
		TOKEN t;
		t= token(s,curr);
		if (!t)
			compile_error();
		if (t->type==tk_lparen) {
			token_destroy(t);
			break;
		}
		if (t->type!=tk_space)
			compile_error();
		token_destroy(t);
	}
	int top= lua_gettop(Lua);
	LIST head;
	{
		int level=1;
		M5STREAM collect= m5stream_open();
		CLL_INIT(&head);
		while(level) {
			TOKEN t;
			t= token(s,curr);
			if ((t->type==tk_comma ||t->type==tk_rparen) && level==1) {
				char *pd;
				{
					M5STREAM parsed= m5stream_open();
					parse(parsed,collect);
					pd= sread_whole(parsed);
					m5stream_close(parsed);
				}
				m5stream_close(collect);
				collect= m5stream_open();
				{
					LIST *n;
					n= malloc(sizeof(*n));
					n->data= pd;
					CLL_INSERT_BEFORE(&head,n);
				}
				if (t->type==tk_comma) {
					token_destroy(t);
					continue;
				}
			}
			if (t->type==tk_rparen) {
				--level;
				if (level==0) {
					m5stream_close(collect);
					token_destroy(t);
					break;
				}
			}
			if (t->type==tk_lparen)
				++level;
			m5stream_copy(collect,t->output);
			token_destroy(t);
		}
	}
	lua_settop(Lua,top);
	{
		LIST *p,*n;
		int argc= 0;
		for(p=head.next;p!=&head;p=n) {
			lua_pushstring(Lua, p->data);
			free(p->data);
			argc++;
			n= p->next;
			free(p);
		}
		if (lua_pcall(Lua,argc, 1,-argc-2)) {
			fprintf(stderr,"while expanding %s,runtime error at %s:%d:\n%s",
						name,
						s->prettyname? s->prettyname : "",lineno+1,
						lua_tostring(Lua,-1));
			lua_pop(Lua,1);
			print_stackc("");
			exit(1);
		} else {
			if (lua_isstring(Lua,-1)) {
				lua_pushvalue(Lua,-1);
				const char *z= lua_tostring(Lua,-1);
				m5stream_write(d, z, strlen(z));
				lua_pop(Lua,1);
			}
			lua_remove(Lua,-2);
		}
	}
	return 1;
}

static void parse(M5STREAM d,M5STREAM s) {
	int curr;
	char c;
	TOKEN t;
	size_t r= sread(s,&c,1);
	if (r<=0)
		curr= EOF;
	else
		curr= c;
	do {
		t= token(s,&curr);
		if (!t)
			break;
		switch(t->type) {
			case tk_error:
				compile_error();
				break;
			case tk_lbegin: {
				int lineno= s->nlines;
				char *pn= s->prettyname;
				token_destroy(t);
				M5STREAM st= m5stream_open();
				while(1) {
					t= token(s,&curr);
					if (!t || t->type==tk_lend)
						break;
					m5stream_copy(st, t->output);
					token_destroy(t);
				}
				char *e= sread_whole(st);
				m5stream_close(st);
				exec_lua(d,e,lineno,pn);
				break;
			}
			case tk_identifier: {
				char *id= sread_whole(t->output);
				if (!parse_exec_macro(d, id, s,&curr))
					m5stream_copy(d, t->output);
				free(id);
				break;
			}
			default: m5stream_copy(d, t->output);
		}
		if (t)
			token_destroy(t);
	} while(1);
}





/* '##K'  PUBLIC  FUNCTIONS */


ssize_t m5read(M5STREAM s,char *buf,size_t len) {
	return sread(s,buf,len); /* hmmm */
}

static void exec_lua(M5STREAM d,char *x,int lineno,char *pn) {
	char buf[2000];
	sprintf(buf,"luablock@(%s:%d)", pn? pn : "", lineno);
	if (luaL_loadbuffer(Lua, x, strlen(x), buf)) {
		const char *errstr= lua_tostring(Lua,-1);
		fprintf(stderr, "%s\n",errstr);
		exit(1);
	}
	int ret= lua_pcall(Lua, 0,1,0);
	if (ret!=0) {
		fprintf(stderr,"%s",lua_tostring(Lua,-1));
		exit(1);
	}
	if (!lua_isnil(Lua,-1)) {
		const char *z= lua_tostring(Lua,-1);
		m5stream_write(d, z, strlen(z));
	}
	lua_settop(Lua,0);
}

static int addmacro(lua_State *L) {
	if (lua_gettop(L)<2 || !lua_isfunction(L,-1) || !lua_isstring(L,-2))
		goto err_usage;
        lua_getglobal(L, "all_macros");
	if (!lua_istable(L,-1))
		goto err_nomacros;
	lua_pushvalue(L, -3);
	lua_pushvalue(L, -3);
	lua_settable(L,-3);
        lua_setglobal(L, "all_macros");
	lua_settop(L,0);
	return 0;
err_usage:
	lua_settop(L,0);
	lua_pushstring(L, "usage: m5macro( \"macroname\" , function )");
	lua_error(L);
	return 0;
err_nomacros:
	lua_settop(L,0);
	lua_pushstring(L,"all_macros is set to nil. aborting addmacro");
	lua_error(L);
	return 0;
}


static void checkstr(lua_State *L) {
	if (lua_gettop(L)<1 || !lua_isstring(L,-1)) {
		lua_pushstring(L,"usage: m5file(<filename>)");
		lua_error(L);
	}
}

static int dofile(lua_State *L) {
	checkstr(L);
	const char *fn= lua_tostring(L,-1);
	int fd= open(fn, O_RDONLY);
	if (fd<0) {
		char buf[2000];
		sprintf(buf,"open %s: %s\n",fn, strerror(errno));
		lua_pop(L,1);
		lua_pushstring(L,buf);
		lua_error(L);
	}
	M5STREAM i= fdstream_open(fd);
	M5STREAM o= m5stream_open();
	i->prettyname= strdup(fn);
	parse(o, i);
	m5stream_close(i);
	close(fd);
	char *r= sread_whole(o);
	m5stream_close(o);
	lua_pop(L,1);
	lua_pushstring(L,r);
	return 1;
}

static int dostring(lua_State *L) {
	checkstr(L);
	const char *str= lua_tostring(L,-1);
	M5STREAM i= memscopen((char*) str);
	M5STREAM o= m5stream_open();
	parse(o, i);
	m5stream_close(i);
	char *r= sread_whole(o);
	m5stream_close(o);
	lua_pop(L,1);
	lua_pushstring(L,r);
	return 1;
}

static int doeval(lua_State *L) {
	checkstr(L);
	const char *expr= lua_tostring(L,-1);
	const char *ret= "return ";
	char *x= malloc(strlen(expr)+ strlen(ret)+1);
	sprintf(x,"%s%s",ret, expr);
	luaL_loadbuffer(L,x, strlen(x), "internal");
	free(x);
	if (lua_pcall(L, 0, 1,0)) {
		fprintf(stderr,"%s\n",lua_tostring(L,-1));
		return 0;
	} else {
		return 1;
	}
}


void shift(int *argc,char **argv) {
	int i;
	for(i=1;i<*argc-1;i++)
		argv[i]= argv[i+1];
	(*argc)--;
}

void usage() {
	errx(3,"usage: m5 [-f input] [-o output] [-h]");
}

void help()
{
puts("Program can use standard input or output if not given on the");
puts("command line. It expands macros found in the input. Each macro");
puts("is defined using some lua code surrounded by ${ }$ pairs.");
puts("");
puts("In order to define a macro, you do this:");
puts(" ${ m5macro( \"square\", function (y) ");
puts("      y= m5eval(y)");
puts("      return y*y");
puts("    end)   }$");
puts("After this, you can use the macro without any special notation:");
puts("   square(5) ");
puts("in the input gets expanded to ");
puts("   25");
puts("");
puts("Here are the meanings of the various lua functions:");
puts("m5file(\"filename\")");
puts("m5str(\"string\")");
puts("	parses the file or the string and returns the expanded");
puts("	contents as a string");
puts("m5quote(string)");
puts("	quotes the string according to C string literal rules");
puts("m5eval(string)");
puts("	evaluates the string as an expression. Useful for converting a ");
puts("        string argument to an integer and such. However, it's not limited");
puts("        to this use.");
puts("");
puts("If you want to use the m5 macros inside the lua code (surrounded by ${ }$ ");
puts("pairs), you can refer to the all_macros table. For the above example:");
puts("");
puts("   ${ x= all_macros.square(3) }$");
puts("");
puts("would work.");
exit(0);
}

int main(int argc,char **argv) {
	Lua= luaL_newstate();
        luaL_openlibs(Lua);
	char *init= " all_macros= {} ; ";
	luaL_loadbuffer(Lua, init, strlen(init), "init");
	lua_pcall(Lua, 0,0,0);
	lua_register(Lua,"m5macro", addmacro);
	lua_register(Lua,"m5file", dofile);
	lua_register(Lua,"m5str", dostring);
	lua_register(Lua,"m5eval", doeval);
	int fd_in= 0;
	int fd_out= 1;
	char *inf="-";
	char *outf="-";
	while(argc>1) {
	   if (!strcmp(argv[1],"-h"))  help();
		if (!strncmp(argv[1],"-f",2)) {
			inf= argv[1]+2;
			if (!*inf) {
				shift(&argc,argv);
				if (argc<2)
					usage();
				inf= argv[1];
			}
			shift(&argc,argv);
			if (strcmp(inf,"-"))
				fd_in= open(inf, O_RDONLY);
			if (fd_in<0)
				err(3,"open %s",inf);
		} else if (!strncmp(argv[1],"-o",2)) {
			outf= argv[1]+2;
			if (!*outf) {
				shift(&argc,argv);
				if (argc<2)
					usage();
				outf= argv[1];
			}
			shift(&argc,argv);
			if (strcmp(outf,"-"))
				fd_in= open(inf, O_WRONLY | O_TRUNC, 0644);
			if (fd_in<0)
				err(3,"open %s",inf);
		} else
			usage();
	}
	{
		M5STREAM in= fdstream_open(fd_in);
		M5STREAM out= fdstream_open(fd_out);
		in->prettyname= inf;
		out->prettyname= outf;
		parse(out,in);
		m5stream_close(in);
		m5stream_close(out);
	}
	close(fd_in);
	close(fd_out);
	return 0;
}

/* LIBRARY FUNCTIONS */
/*

m5file("filename)
m5str("string")
	parses the file or the string and returns the expanded
	contents as a string

m5quote( string )
	quotes the string according to C string literal rules

m5eval(string)
	evaluates the string as an integer, useful when you need the integer
	value of an argument. only defined inside lua blocks


debugging:
if an error occurs during parsing (ie. source-level),
print the location of the error
exit

if an error occurs during parsing a lua-block, from source
print the location of the error (we can only print linenumber of 
the source, then the error message from lua

if an error occurs during execution of lua code,
find the caller's location and print it.
get the stack info and dump it.


*/
